﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.ServiceModel;
using System.ServiceProcess;
using WpkgUtilities;

namespace WpkgInstaller
{
    public enum ServiceState
    {
        SERVICE_STOPPED = 0x00000001,
        SERVICE_START_PENDING = 0x00000002,
        SERVICE_STOP_PENDING = 0x00000003,
        SERVICE_RUNNING = 0x00000004,
        SERVICE_CONTINUE_PENDING = 0x00000005,
        SERVICE_PAUSE_PENDING = 0x00000006,
        SERVICE_PAUSED = 0x00000007,
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct ServiceStatus
    {
        public int dwServiceType;
        public ServiceState dwCurrentState;
        public int dwControlsAccepted;
        public int dwWin32ExitCode;
        public int dwServiceSpecificExitCode;
        public int dwCheckPoint;
        public int dwWaitHint;
    };

    [ServiceBehavior(InstanceContextMode = InstanceContextMode.Single)]
    public partial class WpkgInstaller : ServiceBase, IWpkgNotifierMessageServer
    {

        [DllImport("advapi32.dll", SetLastError = true)]
        private static extern bool SetServiceStatus(System.IntPtr handle, ref ServiceStatus serviceStatus);
        private Dictionary<int, IWpkgNotifierMessageClient> clients;

        public WpkgInstaller()
        {
            InitializeComponent();
            
            eventLog = new EventLog();
            clients = new Dictionary<int, IWpkgNotifierMessageClient>();
            string eventSourceName = "WpkgInstaller Service";
            string logName = "WpkgNotifier";

            if (!EventLog.SourceExists(eventSourceName))
            {
                EventLog.CreateEventSource(eventSourceName, logName);
            }

            eventLog.Source = eventSourceName;
            eventLog.Log = logName;

            OpenMessageChannel();

        }

        protected override void OnStart(string[] args)
        {

            ServiceStatus serviceStatus = new ServiceStatus
            {
                dwCurrentState = ServiceState.SERVICE_START_PENDING,
                dwWaitHint = 100000
            };
            SetServiceStatus(ServiceHandle, ref serviceStatus);
            // RunWpkgSync();
            serviceStatus.dwCurrentState = ServiceState.SERVICE_RUNNING;
            SetServiceStatus(ServiceHandle, ref serviceStatus);

            eventLog.WriteEntry("Service started", EventLogEntryType.Information);
        }

        protected override void OnStop()
        {
            ServiceStatus serviceStatus = new ServiceStatus
            {
                dwCurrentState = ServiceState.SERVICE_STOP_PENDING,
                dwWaitHint = 100000
            };
            SetServiceStatus(ServiceHandle, ref serviceStatus);

            serviceStatus.dwCurrentState = ServiceState.SERVICE_STOPPED;
            SetServiceStatus(ServiceHandle, ref serviceStatus);

            eventLog.WriteEntry("Service stopped", EventLogEntryType.Information);
        }

        private void RunWpkgSync()
        {
            eventLog.WriteEntry("Starting sync", EventLogEntryType.Information);
            RunWpkgCommand("/synchronize");
        }

        private void RunWpkgQuery()
        {
            eventLog.WriteEntry("Starting query", EventLogEntryType.Information);
            RunWpkgCommand("/query:m");
        }
        private void RunWpkgCommand(string arg)
        {
            Process process = new Process();
            ProcessStartInfo startInfo = new ProcessStartInfo()
            {
                RedirectStandardInput = true,
                RedirectStandardOutput = true,
                RedirectStandardError = true,
                WindowStyle = ProcessWindowStyle.Hidden,

                FileName = Environment.GetEnvironmentVariable("comspec"),
                Arguments = String.Format("/C \\\\wpkg.ch.cam.ac.uk\\wpkg\\wpkg.cmd {0}", arg),
                WorkingDirectory = Environment.GetEnvironmentVariable("temp"),

                UseShellExecute = false,
            };

            process.StartInfo = startInfo;

            process.EnableRaisingEvents = true;
            process.OutputDataReceived += ProcessCommandStdout;
            process.ErrorDataReceived += ProcessCommandStderr;

            process.Start();
            process.BeginErrorReadLine();
            process.BeginOutputReadLine();
            process.WaitForExit();

            if(process.ExitCode == 0)
            {
                SetClientsState(WpkgMessageState.SUCCESS);
            }
            else
            {
                SetClientsState(WpkgMessageState.ERROR);
            }
            String msg = String.Format("wpkg command finished, with exit code {0}", process.ExitCode);
            eventLog.WriteEntry(msg, EventLogEntryType.Information);

        }

        private void ProcessCommandStdout(object sendingProcess, DataReceivedEventArgs e)
        {
            if (e != null && e.Data != null)
            {
                SendMessageToClients(e.Data);
                // sent e.Data back to clients
            }

        }

        private void ProcessCommandStderr(object sendingProcess, DataReceivedEventArgs e)
        {
            if (e != null && e.Data != null)
            {
                SendMessageToClients(e.Data);
                eventLog.WriteEntry(e.Data, EventLogEntryType.Error);
            }

        }

        private void eventLog_EntryWritten(object sender, EntryWrittenEventArgs e)
        {

        }

        private void OpenMessageChannel()
        {
            try
            {
                Uri uri = new Uri("net.pipe://localhost/www.ch.cam.ac.uk/WpkgNotifier");
                NetNamedPipeBinding binding = new NetNamedPipeBinding();
                ServiceHost serviceHost = new ServiceHost(this, uri);
                serviceHost.AddServiceEndpoint(typeof(IWpkgNotifierMessageServer), binding, "");

                serviceHost.Open();

                eventLog.WriteEntry("serviceHost now open");

            } 
            catch (Exception e)
            {
                eventLog.WriteEntry(e.Message, EventLogEntryType.Error);
                throw e;
            }

        }

        public void Register(int pid)
        {
            eventLog.WriteEntry(String.Format("Register requested for {0}", pid));
            IWpkgNotifierMessageClient client = OperationContext.Current.GetCallbackChannel <IWpkgNotifierMessageClient>();
            clients.Add(pid, client);
            
            foreach(KeyValuePair<int, IWpkgNotifierMessageClient> kv in clients)
            {
                try
                {
                    kv.Value.StatusLine(String.Format("Register requested for {0}", pid));
                }
                catch (Exception e)
                {
                    eventLog.WriteEntry("Error communicating with client " + kv.Key + " " + e.ToString(), EventLogEntryType.Warning);
                }
            }
        }

        public void Unregister(int pid)
        {
            if(clients.Remove(pid))
            {
                eventLog.WriteEntry(String.Format("Client {0} unregistered", pid));
            }
            else
            {
                eventLog.WriteEntry(String.Format("Unknown client {0} requested to unregister", pid), EventLogEntryType.Warning);
            }
        }
        public void ProcessRequest(WpkgMessageRequestType type)
        {
            switch (type)
            {
                case WpkgMessageRequestType.DO_SYNC:
                    SetClientsState(WpkgMessageState.SYNCING);
                    RunWpkgSync();
                    break;
                case WpkgMessageRequestType.RUN_QUERY:
                    SetClientsState(WpkgMessageState.QUERYING);
                    RunWpkgQuery();
                    break;

            }
        }

        private void SetClientsState(WpkgMessageState state)
        {
            foreach (KeyValuePair<int, IWpkgNotifierMessageClient> kv in clients)
            {
                try
                {
                    kv.Value.SetStatus(state);
                }
                catch (Exception e)
                {
                    eventLog.WriteEntry("Error communicating with client " + kv.Key + " " + e.ToString(), EventLogEntryType.Warning);
                }
            }
        }

        private void SendMessageToClients(string msg)
        {
            foreach (KeyValuePair<int, IWpkgNotifierMessageClient> kv in clients)
            {
                try
                {
                    kv.Value.StatusLine(msg);
                }
                catch (Exception e)
                {
                    eventLog.WriteEntry("Error communicating with client " + kv.Key + " " + e.ToString(), EventLogEntryType.Warning);
                }
            }
        }
    }
}
